// SPDX-License-Identifier: GPL-2.0-or-later

/*
 * Copyright (C) 2025 Alibaba Inc
 */

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <traceevent/trace-seq.h>
#include <unistd.h>

#include "cper.h"
#include "ras-logger.h"
#include "ras-bert.h"
#include "ras-record.h"
#include "ras-mc-handler.h"
#include "ras-aer-handler.h"

#define BERT_HANDLED 	"/run/rasdaemon/bert"
#define BERT_DATA 		"/sys/firmware/acpi/tables/data/BERT"
#define BERT_TABLE 		"/sys/firmware/acpi/tables/BERT"

#define ACPI_BERT_PRINT_MAX_RECORDS 5
#define ACPI_BERT_PRINT_MAX_LEN 1024

static bool bert_handled(void)
{
	return access(BERT_HANDLED, F_OK) == 0;
}

static bool bert_exist(void)
{
	return access(BERT_DATA, F_OK) == 0;
}

static void bert_handled_tag(void)
{
    FILE *file = fopen(BERT_HANDLED, "w");
    if (file == NULL) {
        log(SYSLOG, LOG_ERR, "Failed to create BERT tag file");
        return;
    }

    fclose(file);
}

static unsigned bcd2bin(unsigned char val)
{
	return (val & 0x0f) + (val >> 4) * 10;
}

static void cper_print_tstamp(char *t,
				   struct acpi_hest_generic_data_v300 *gdata)
{
	uint8_t hour, min, sec, day, mon, year, century, *timestamp;

	if (gdata->validation_bits & ACPI_HEST_GEN_VALID_TIMESTAMP) {
		timestamp = (uint8_t *)&(gdata->time_stamp);
		sec       = bcd2bin(timestamp[0]);
		min       = bcd2bin(timestamp[1]);
		hour      = bcd2bin(timestamp[2]);
		day       = bcd2bin(timestamp[4]);
		mon       = bcd2bin(timestamp[5]);
		year      = bcd2bin(timestamp[6]);
		century   = bcd2bin(timestamp[7]);

		snprintf(t, 64, 
			"%02d%02d-%02d-%02d %02d:%02d:%02d",
					century, year, mon, day, hour, min, sec);
	}
}

static int cper_estatus_check_header(const struct acpi_hest_generic_status *estatus)
{
	if (estatus->data_length &&
	    estatus->data_length < sizeof(struct acpi_hest_generic_data))
		return -EINVAL;
	if (estatus->raw_data_length &&
	    estatus->raw_data_offset < sizeof(*estatus) + estatus->data_length)
		return -EINVAL;

	return 0;
}

static int cper_estatus_check(const struct acpi_hest_generic_status *estatus)
{
	struct acpi_hest_generic_data *gdata;
	unsigned int data_len, record_size;
	int rc;

	rc = cper_estatus_check_header(estatus);
	if (rc)
		return rc;

	data_len = estatus->data_length;

	apei_estatus_for_each_section(estatus, gdata) {
		if (acpi_hest_get_size(gdata) > data_len)
			return -EINVAL;

		record_size = acpi_hest_get_record_size(gdata);
		if (record_size > data_len)
			return -EINVAL;

		data_len -= record_size;
	}
	if (data_len)
		return -EINVAL;

	return 0;
}

static int cper_mem_err_location(struct cper_sec_mem_err *mem, char *msg)
{
	uint32_t len, n;

	if (!msg)
		return 0;

	n = 0;
	len = CPER_REC_LEN;
	if (mem->validation_bits & CPER_MEM_VALID_NODE)
		n += snprintf(msg + n, len - n, "node:%d ", mem->node);
	if (mem->validation_bits & CPER_MEM_VALID_CARD)
		n += snprintf(msg + n, len - n, "card:%d ", mem->card);
	if (mem->validation_bits & CPER_MEM_VALID_MODULE)
		n += snprintf(msg + n, len - n, "module:%d ", mem->module);
	if (mem->validation_bits & CPER_MEM_VALID_RANK_NUMBER)
		n += snprintf(msg + n, len - n, "rank:%d ", mem->rank);
	if (mem->validation_bits & CPER_MEM_VALID_BANK)
		n += snprintf(msg + n, len - n, "bank:%d ", mem->bank);
	if (mem->validation_bits & CPER_MEM_VALID_BANK_GROUP)
		n += snprintf(msg + n, len - n, "bank_group:%d ",
			       mem->bank >> CPER_MEM_BANK_GROUP_SHIFT);
	if (mem->validation_bits & CPER_MEM_VALID_BANK_ADDRESS)
		n += snprintf(msg + n, len - n, "bank_address:%d ",
			       mem->bank & CPER_MEM_BANK_ADDRESS_MASK);
	if (mem->validation_bits & CPER_MEM_VALID_DEVICE)
		n += snprintf(msg + n, len - n, "device:%d ", mem->device);
	if (mem->validation_bits & (CPER_MEM_VALID_ROW | CPER_MEM_VALID_ROW_EXT)) {
		uint32_t row = mem->row;

		row |= cper_get_mem_extension(mem->validation_bits, mem->extended);
		n += snprintf(msg + n, len - n, "row:%d ", row);
	}
	if (mem->validation_bits & CPER_MEM_VALID_COLUMN)
		n += snprintf(msg + n, len - n, "column:%d ", mem->column);
	if (mem->validation_bits & CPER_MEM_VALID_BIT_POSITION)
		n += snprintf(msg + n, len - n, "bit_position:%d ",
			       mem->bit_pos);
	if (mem->validation_bits & CPER_MEM_VALID_REQUESTOR_ID)
		n += snprintf(msg + n, len - n, "requestor_id:0x%016lx ",
			       mem->requestor_id);
	if (mem->validation_bits & CPER_MEM_VALID_RESPONDER_ID)
		n += snprintf(msg + n, len - n, "responder_id:0x%016lx ",
			       mem->responder_id);
	if (mem->validation_bits & CPER_MEM_VALID_TARGET_ID)
		n += snprintf(msg + n, len - n, "target_id:0x%016lx ",
			       mem->target_id);
	if (mem->validation_bits & CPER_MEM_VALID_CHIP_ID)
		n += snprintf(msg + n, len - n, "chip_id:%d ",
			       mem->extended >> CPER_MEM_CHIP_ID_SHIFT);

	return n;
}

static const char *cper_mem_err_status_str(uint64_t status)
{
	switch ((status >> 8) & 0xff) {
	case  1:	return "Error detected internal to the component";
	case  4:	return "Storage error in DRAM memory";
	case  5:	return "Storage error in TLB";
	case  6:	return "Storage error in cache";
	case  7:	return "Error in one or more functional units";
	case  8:	return "Component failed self test";
	case  9:	return "Overflow or undervalue of internal queue";
	case 16:	return "Error detected in the bus";
	case 17:	return "Virtual address not found on IO-TLB or IO-PDIR";
	case 18:	return "Improper access error";
	case 19:	return "Access to a memory address which is not mapped to any component";
	case 20:	return "Loss of Lockstep";
	case 21:	return "Response not associated with a request";
	case 22:	return "Bus parity error - must also set the A, C, or D Bits";
	case 23:	return "Detection of a protocol error";
	case 24:	return "Detection of a PATH_ERROR";
	case 25:	return "Bus operation timeout";
	case 26:	return "A read was issued to data that has been poisoned";
	default:	return "Reserved";
	}
}

static int print_mem_error_other_detail(const struct cper_sec_mem_err *mem, char *msg,
					const char *location, unsigned int len)
{
	uint32_t n;

	if (!msg)
		return 0;

	n = 0;
	len -= 1;

	n += snprintf(msg + n, len - n, "APEI location: %s ", location);

	if (!(mem->validation_bits & CPER_MEM_VALID_ERROR_STATUS))
		goto out;

	n += snprintf(msg + n, len - n, "status(0x%016lx): ", mem->error_status);
	n += snprintf(msg + n, len - n, "%s ", cper_mem_err_status_str(mem->error_status));

out:
	msg[n] = '\0';

	return n;
}

static const char *cper_mem_err_type_str(unsigned int etype)
{
	return etype < ARRAY_SIZE(mem_err_type_strs) ?
		mem_err_type_strs[etype] : "unknown";
}

static void cper_print_mem_error(struct acpi_hest_generic_data *gdata)
{ 
		struct cper_sec_mem_err *mem_err = acpi_hest_get_payload(gdata);
		struct ras_mc_event e = { 0 };
		char location[256];
		char other_detail[400];
		char msg[80];
		char *p;
		const guid_t *fru_id = NULL;
		int sec_sev;
		char *fru_text = "";
		struct trace_seq s;
		size_t len;

		sec_sev = ghes_severity(gdata->error_severity);
		if (gdata->validation_bits & CPER_SEC_VALID_FRU_ID)
			fru_id = (guid_t *)gdata->fru_id;

		if (gdata->validation_bits & CPER_SEC_VALID_FRU_TEXT)
			fru_text = (char *)gdata->fru_text;

		cper_print_tstamp(e.timestamp, (struct acpi_hest_generic_data_v300 *)gdata);
		if (!e.timestamp[0]) {
			time_t now;
			struct tm *tm;

			now = time(NULL);

			tm = localtime(&now);
			if (tm)
				strftime(e.timestamp, sizeof(e.timestamp),
					"%Y-%m-%d %H:%M:%S %z", tm);
		}

		e.label = "unknown memory";
		e.severity = sec_sev;
		e.error_count = 1;
		e.grain = 1;
		e.msg = msg;
		e.driver_detail = other_detail;
		e.top_layer = -1;
		e.middle_layer = -1;
		e.lower_layer = -1;
		*other_detail = '\0';
		*msg = '\0';
		switch (sec_sev) {
		case GHES_SEV_CORRECTED:
			e.error_type = "Corrected";
			break;
		case GHES_SEV_RECOVERABLE:
			e.error_type = "Uncorrected";
			break;
		case GHES_SEV_PANIC:
			e.error_type = "Fatal";
			break;
		default:
		case GHES_SEV_NO:
			e.error_type = "Info";
		}

		if (mem_err->validation_bits & CPER_MEM_VALID_ERROR_TYPE) {
			uint8_t etype = mem_err->error_type;

			p = msg;
			p += snprintf(p, sizeof(msg), "%s", cper_mem_err_type_str(etype));
		} else {
			strcpy(msg, "unknown error");
		}
		e.msg = msg;

		if (mem_err->validation_bits & CPER_MEM_VALID_PA)
			e.address =  mem_err->physical_addr;
	
		if (mem_err->validation_bits & CPER_MEM_VALID_PA_MASK)
			e.grain = ~mem_err->physical_addr_mask + 1;

		p = location;
		p += cper_mem_err_location(mem_err, p);

		len = p - location;
		if (len)
			location[len - 1] = '\0';
		
		/* All other fields are mapped on e->other_detail */
		p = other_detail;
		if (fru_id)
			p += snprintf(p, 30, "fru_id: %pUl ", fru_id);
		if (fru_text)
			p += snprintf(p, 30, "fru_text: %.20s ", fru_text);
		p += print_mem_error_other_detail(mem_err, p, location, 400);
		
		if (p > other_detail)
			*(p - 1) = '\0';
		e.driver_detail = other_detail;

		trace_seq_init(&s);
		trace_seq_printf(&s, "%16s-%-10d [%03d] %s %6.6f %25s: ",
				"<...>", 0, -1, "....", 0.0f, BERT_MEM_EVENT_NAME);

		report_mc_event(&e, &s);
		trace_seq_terminate(&s);
		trace_seq_do_printf(&s);
		printf("\n");
		fflush(stdout);
		trace_seq_destroy(&s);
}

static int cper_severity_to_aer(int cper_severity)
{
	switch (cper_severity) {
	case CPER_SEV_RECOVERABLE:
		return AER_NONFATAL;
	case CPER_SEV_FATAL:
		return AER_FATAL;
	default:
		return AER_CORRECTABLE;
	}
}

static void cper_print_pcie_aer(struct acpi_hest_generic_data *gdata)
{
	struct cper_sec_pcie *pcie_err = acpi_hest_get_payload(gdata);
	struct ras_aer_event e = { 0 };
	unsigned int devfn;
	int aer_severity, tlp_header_valid = 0;
	struct aer_capability_regs aer_info;
	uint32_t status, mask;
	struct trace_seq s;
	char buf[80];
	uint32_t tlp_header[4];

	if (!(pcie_err->validation_bits & CPER_PCIE_VALID_DEVICE_ID) ||
	    !(pcie_err->validation_bits & CPER_PCIE_VALID_AER_INFO))
		return;

	cper_print_tstamp(e.timestamp, (struct acpi_hest_generic_data_v300 *)gdata);
	if (!e.timestamp[0]) {
		time_t now;
		struct tm *tm;

		now = time(NULL);

		tm = localtime(&now);
		if (tm)
			strftime(e.timestamp, sizeof(e.timestamp),
				"%Y-%m-%d %H:%M:%S %z", tm);
	}

	devfn = PCI_DEVFN(pcie_err->device_id.device,
				pcie_err->device_id.function);
	aer_severity = cper_severity_to_aer(gdata->error_severity);

	/*
		* If firmware reset the component to contain
		* the error, we must reinitialize it before
		* use, so treat it as a fatal AER error.
		*/
	if (gdata->flags & CPER_SEC_RESET)
		aer_severity = AER_FATAL;

	memcpy(&aer_info, pcie_err->aer_info, sizeof(struct aer_capability_regs));

	snprintf(buf, 80, "%.4x:%.2x:%.2x.%x", 
										pcie_err->device_id.segment, 
										pcie_err->device_id.bus, 
										PCI_SLOT(devfn), 
										PCI_FUNC(devfn));
	e.dev_name = buf;
	e.severity = aer_severity;

	if (aer_severity == AER_CORRECTABLE) {
		status = aer_info.cor_status;
		mask = aer_info.cor_mask;
	} else {
		status = aer_info.uncor_status;
		mask = aer_info.uncor_mask;
		tlp_header_valid = status & AER_LOG_TLP_MASKS;
	}

	e.status = (status & ~mask);
	e.tlp_header_valid = tlp_header_valid;
	tlp_header[0] = aer_info.header_log.dw[0];
	tlp_header[1] = aer_info.header_log.dw[1];
	tlp_header[2] = aer_info.header_log.dw[2];
	tlp_header[3] = aer_info.header_log.dw[3];
	e.tlp_header = tlp_header;

	trace_seq_init(&s);
	trace_seq_printf(&s, "%16s-%-10d [%03d] %s %6.6f %25s: ",
			"<...>", 0, -1, "....", 0.0f, BERT_AER_EVENT_NAME);

	report_aer_event(&s, &e);
	trace_seq_terminate(&s);
	trace_seq_do_printf(&s);
	printf("\n");
	fflush(stdout);
	trace_seq_destroy(&s);
}

static void cper_print_bits(unsigned int bits,
		     const char * const strs[], unsigned int strs_size, struct trace_seq *s)
{
	int i, len = 0;
	const char *str;
	char buf[84];

	for (i = 0; i < strs_size; i++) {
		if (!(bits & (1U << i)))
			continue;
		str = strs[i];
		if (!str)
			continue;
		if (!len)
			len = snprintf(buf, sizeof(buf), "%s", str);
		else
			len += snprintf(buf+len, sizeof(buf)-len, ", %s", str);
	}
	if (len)
		trace_seq_printf(s, "%s ", buf);
}

static void cper_print_proc_generic(struct acpi_hest_generic_data *gdata)
{
	struct cper_sec_proc_generic *proc = acpi_hest_get_payload(gdata);
	struct trace_seq s;
	char timestamp[64];

	trace_seq_init(&s);
	trace_seq_printf(&s, "%16s-%-10d [%03d] %s %6.6f %25s: ",
			"<...>", 0, -1, "....", 0.0f, BERT_AER_EVENT_NAME);

	cper_print_tstamp(timestamp, (struct acpi_hest_generic_data_v300 *)gdata);
	if (!timestamp[0]) {
		time_t now;
		struct tm *tm;

		now = time(NULL);

		tm = localtime(&now);
		if (tm)
			strftime(timestamp, sizeof(timestamp),
				"%Y-%m-%d %H:%M:%S %z", tm);
	}
	trace_seq_printf(&s, "%s ", timestamp);

	if (proc->validation_bits & CPER_PROC_VALID_TYPE)
		trace_seq_printf(&s, "processor_type: %s ", proc->proc_type < ARRAY_SIZE(proc_type_strs) ?
		       proc_type_strs[proc->proc_type] : "unknown");
	if (proc->validation_bits & CPER_PROC_VALID_ISA)
		trace_seq_printf(&s, "processor_isa: %s ", proc->proc_isa < ARRAY_SIZE(proc_isa_strs) ?
		       proc_isa_strs[proc->proc_isa] : "unknown");
	if (proc->validation_bits & CPER_PROC_VALID_ERROR_TYPE) {
		trace_seq_printf(&s, "error_type: 0x%02x ", proc->proc_error_type);
		cper_print_bits(proc->proc_error_type,
				cper_proc_error_type_strs,
				ARRAY_SIZE(cper_proc_error_type_strs), &s);
	}
	if (proc->validation_bits & CPER_PROC_VALID_OPERATION)
		trace_seq_printf(&s, "operation: %s ", proc->operation < ARRAY_SIZE(proc_op_strs) ?
		       proc_op_strs[proc->operation] : "unknown");
	if (proc->validation_bits & CPER_PROC_VALID_FLAGS) {
		trace_seq_printf(&s, "flags: 0x%02x ", proc->flags);
		cper_print_bits(proc->flags, proc_flag_strs,
				ARRAY_SIZE(proc_flag_strs), &s);
	}
	if (proc->validation_bits & CPER_PROC_VALID_LEVEL)
		trace_seq_printf(&s, "level: %d ", proc->level);
	if (proc->validation_bits & CPER_PROC_VALID_VERSION)
		trace_seq_printf(&s, "version_info: 0x%016lx ", proc->cpu_version);
	if (proc->validation_bits & CPER_PROC_VALID_ID)
		trace_seq_printf(&s, "processor_id: 0x%016lx ", proc->proc_id);
	if (proc->validation_bits & CPER_PROC_VALID_TARGET_ADDRESS)
		trace_seq_printf(&s, "target_address: 0x%016lx ", proc->target_addr);
	if (proc->validation_bits & CPER_PROC_VALID_REQUESTOR_ID)
		trace_seq_printf(&s, "requestor_id: 0x%016lx ", proc->requestor_id);
	if (proc->validation_bits & CPER_PROC_VALID_RESPONDER_ID)
		trace_seq_printf(&s, "responder_id: 0x%016lx ", proc->responder_id);
	if (proc->validation_bits & CPER_PROC_VALID_IP)
		trace_seq_printf(&s, "IP: 0x%016lx ", proc->ip);
}

static void
cper_estatus_print_section(struct acpi_hest_generic_data *gdata)
{
	guid_t *sec_type;

	sec_type = (guid_t *)gdata->section_type;
	if (guid_equal(sec_type, &CPER_SEC_PLATFORM_MEM)) {
		cper_print_mem_error(gdata);
	} else if (guid_equal(sec_type, &CPER_SEC_PCIE)) {
		cper_print_pcie_aer(gdata);
	} else if (guid_equal(sec_type, &CPER_SEC_PROC_GENERIC)) {
		cper_print_proc_generic(gdata);
	} else {
		// TODO
	}
}

static void bert_print_all(struct acpi_hest_generic_status *region,
				  unsigned int region_len)
{
	struct acpi_hest_generic_status *estatus = region;
	int remain = region_len;
	uint32_t estatus_len;

	while (remain >= sizeof(struct acpi_hest_generic_status)) {
		struct acpi_hest_generic_data *gdata;

		estatus_len = cper_estatus_len(estatus);
		if (remain < estatus_len) {
			log(SYSLOG, LOG_ERR, "Truncated status block (length: %u).\n",
				estatus_len);
			break;
		}

		if (cper_estatus_check(estatus)) {
			log(SYSLOG, LOG_ERR, "Invalid error record.\n");
			break;
		}

		apei_estatus_for_each_section(estatus, gdata) {
			cper_estatus_print_section(gdata);
		}

		estatus = (void *)estatus + estatus_len;
		remain -= estatus_len;
	}
}

static void handle_bert_cper(void)
{
    FILE *bert_file, *bert_data;
    struct acpi_table_bert bert_table;
	uint32_t bert_length;
	void *bert_buffer;
	
	bert_file = fopen(BERT_TABLE, "rb");
    if (!bert_file) {
        log(SYSLOG, LOG_ERR, "Failed to open BERT table file");
        return;
    }

    if (fread(&bert_table, sizeof(bert_table), 1, bert_file) != 1) {
        log(SYSLOG, LOG_ERR, "Failed to read BERT header\n");
		goto free_file;
    }

	bert_length = bert_table.region_length;

	log(SYSLOG, LOG_INFO, "Find BERT table address: %#lx, length %d\n", 
		bert_table.address, bert_table.region_length);
    bert_data = fopen(BERT_DATA, "rb");
    if (!bert_data){
        log(SYSLOG, LOG_ERR, "Failed to open data file\n");
		goto free_file;
	}

    bert_buffer = malloc(bert_length);
    if (!bert_buffer) {
        log(SYSLOG, LOG_ERR, "Memory allocation failed\n");
		goto free_data;
    }

    if (fread(bert_buffer, 1, bert_length, bert_data) != bert_length) {
        log(SYSLOG, LOG_ERR, "Failed to read full region from data file\n");
		goto free_buf;
    }

	bert_print_all(bert_buffer, bert_length);

free_buf:
	free(bert_buffer);
free_data:
	fclose(bert_data);
free_file:
	fclose(bert_file);

    return;
}

void handle_bert(void)
{
	if (!bert_exist())
		return;

	if (bert_handled())
		return;

	handle_bert_cper();

	bert_handled_tag();
}